
from dataclasses import dataclass, _MISSING_TYPE, fields
import os
from typing import Any, Optional, get_type_hints


class EnvVarPopulater:
    _converts = {
        int: lambda d: int(d),
        float: lambda d: float(d),
        Optional[int]: lambda d: int(d) if d else None,
        Optional[float]: lambda d: float(d) if d else None,
    }

    @classmethod
    def populate(cls) -> Any:
        values = {}
        typehints = get_type_hints(cls)
        for field in fields(cls):
            conv = cls._converts[typehints[field.name]]
            valstr = os.environ.get(field.name.upper())
            if valstr is not None:
                try:
                    values[field.name] = conv(valstr)
                except Exception as ex:
                    raise Exception(f'Error converting environment variable "{field.name.upper()}" as {field.type}') from ex
            else:
                if isinstance(field.default, _MISSING_TYPE) and isinstance(field.default_factory, _MISSING_TYPE):
                    # not provided and no default value
                    raise Exception(f'Environment variable "{field.name.upper()}" is required')
        return cls(** values)


@dataclass
class BatchingConfig(EnvVarPopulater):
    pyraisdk_max_batch_size: int
    pyraisdk_idle_batch_size: int
    pyraisdk_max_batch_interval: float




import multiprocessing as mp
import queue
import sys
import signal
import threading
import time
import traceback

from typing import List, Optional

from .log import (
    LOGGING_QUEUE_CAPACITY,
    SINK_BATCH_MAX_IDLE_SECS,
    Event,
    EventLogger,
    EventSink,
    LogLevel,
    UnstructuredEvent,
    StructuredEvent,
    TerminateEvent,
    ERR_PREFIX,
)


class AsyncEventLogger(EventLogger):
    def __init__(self, role: str = '', instance: str=''):
        super(AsyncEventLogger, self).__init__()
        self.role = role 
        self.instance = instance
        self.unstruct_queue: mp.Queue[Event] = mp.Queue(maxsize=LOGGING_QUEUE_CAPACITY)
        self.struct_queue: mp.Queue[Event] = mp.Queue(maxsize=LOGGING_QUEUE_CAPACITY)
        self.backend_loop: Optional[_AsyncEventLoggerBackendLoop] = None
    
    def log(self, level: LogLevel, msg: str, corr_id: str='', elem: int=-1):
        # check logger started
        if self.backend_loop is None:
            return
        evt = UnstructuredEvent(
                level = level,
                role = self.role,
                instance = self.instance,
                context = '',
                message = msg,
                correlation_id = corr_id,
                element = elem,
            )
        try:
            self.unstruct_queue.put_nowait(evt)
        except queue.Full:
            msg = f'{ERR_PREFIX} AsyncEventLogger shedding an event. Unstruct buffer grew faster than stream could be written.'
            print(msg, file=sys.stderr)

    def event(self, key: str, code: str, numeric: float, detail: str='', corr_id: str='', elem: int=-1):
        # check logger started
        if self.backend_loop is None:
            return
        evt = StructuredEvent(
            key = key,
            role = self.role,
            instance = self.instance,
            correlation_id = corr_id,
            element = elem,
            code = code,
            numeric = numeric,
            detail = detail,
        )
        try:
            self.struct_queue.put_nowait(evt)
        except queue.Full:
            msg = f'{ERR_PREFIX} AsyncEventLogger shedding an event. Struct buffer grew faster than stream could be written.'
            print(msg, file=sys.stderr)

    def start(self):
        with self.mutex:
            if self.backend_loop:
                return
            self.backend_loop = _AsyncEventLoggerBackendLoop(
                self.unstruct_queue,
                self.unstruct_sinks,
                self.struct_queue,
                self.struct_sinks
            )
            self.backend_loop.start()
    
    def close(self):
        with self.mutex:
            if self.backend_loop is not None:
                self.backend_loop.close()

    def __del__(self):
        self.close()


class _AsyncEventLoggerBackendLoop:
    def __init__(
        self,
        unstruct_queue: mp.Queue,
        unstruct_sinks: List[EventSink],
        struct_queue: mp.Queue,
        struct_sinks: List[EventSink],
    ):
        self.categories_q: List[mp.Queue[Event]] = [unstruct_queue, struct_queue]
        self.categories_sinks: List[List[EventSink]] = [unstruct_sinks, struct_sinks]
        self.categories_sinks_ready: Optional[List[List[EventSink]]] = None
        self.terminated = mp.Event()
        self.mutex = mp.RLock()
        self.process: Optional[mp.Process] = None
        self.loop_threads: List[threading.Thread] = []
        
    def start(self):
        with self.mutex:
            if self.process:
                return
            self.process = mp.Process(target=self.main, daemon=True)
            self.process.start()
        
    def main(self):
        ''' entrance of backend process
        '''
        with self.mutex:
            # set quit signal
            signal.signal(signal.SIGINT, self.handle_signal)
            signal.signal(signal.SIGTERM, self.handle_signal)
            # start sinks
            self.categories_sinks_ready = [
                self._get_ready_sinks(sinks) for sinks in self.categories_sinks
            ]
            # start threads
            for q, sinks in zip(self.categories_q, self.categories_sinks_ready):
                t = threading.Thread(target=self.run_q, args=(q, sinks), daemon=True)
                t.start()
                self.loop_threads.append(t)

        for t in self.loop_threads:
            t.join()
        
        # close sinks at the end (not exit by handle_signal)
        for sinks in self.categories_sinks_ready:
            for sink in sinks:
                sink.close()

    def _get_ready_sinks(self, sinks: List[EventSink]) -> List[EventSink]:
        result = []
        for sink in sinks:
            try:
                sink.start()
                result.append(sink)
            except Exception:
                msg = ERR_PREFIX + ' AsyncEventLogger failed to start a sink. Removed it from sink list of logger\n' + traceback.format_exc()
                print(msg, file=sys.stderr)
        return result

    def run_q(self, q: mp.Queue, sinks: List[EventSink]):
        ''' If run_q_inner failed somehow, sleep 3 seconds and retry
            until process terminated
        '''
        while True:
            try:
                self.run_q_inner(q, sinks)
            except Exception:
                traceback.print_exc()

            if self.terminated.is_set():
                break
            else:
                time.sleep(3)

    def run_q_inner(self, q: mp.Queue, sinks: List[EventSink]):
        ''' Fetch events from q, and put them into sinks
        '''
        while not self.terminated.is_set():
            try:
                evt = q.get(block=True, timeout=SINK_BATCH_MAX_IDLE_SECS)
                if isinstance(evt, TerminateEvent):
                    continue
            except queue.Empty:
                continue
            for sink in sinks:
                try:
                    sink.sink(evt)
                except Exception:
                    msg = ERR_PREFIX + f' AsyncEventLogger failed to call sink of {type(sink)}.\n' + traceback.format_exc()
                    print(msg, file=sys.stderr)

        # terminated, flush all
        while True:
            try:
                # Set a small timeout here, to ensure `q` is really consumed up on exiting.
                # To avoid the deadlock issue occuring with a low probability.
                evt = q.get(block=True, timeout=0.1)
                if isinstance(evt, TerminateEvent):
                    continue
                for sink in sinks:
                    try:
                        sink.sink(evt)
                    except Exception:
                        msg = ERR_PREFIX + f' AsyncEventLogger failed to call sink of {type(sink)}.\n' + traceback.format_exc()
                        print(msg, file=sys.stderr)

            except queue.Empty:
                break

    def handle_signal(self, signum, frame):
        with self.mutex:
            self.terminated.set()
            for q in self.categories_q:
                try:
                    q.put_nowait(TerminateEvent())
                except queue.Full:
                    pass
            for t in self.loop_threads:
                t.join()
            if self.categories_sinks_ready:
                for sinks in self.categories_sinks_ready:
                    for sink in sinks:
                        sink.close()

    def close(self):
        with self.mutex:
            if self.terminated.is_set():
                return
            self.terminated.set()
            for q in self.categories_q:
                try:
                    q.put_nowait(TerminateEvent())
                except queue.Full:
                    pass


    def __del__(self):
        self.close()


class StdoutSink(EventSink):
    def sink(self, evt: Event):
        msg = evt.friendly_string()
        sys.stdout.write(msg)




from typing import List, Optional
import psutil
import time

from threading import Thread

from .log import EventLogger
from .gputil import get_gpu_summary


EVENT_KEY_PREFIX = 'pyraisys'


class Avarager:
    def __init__(self):
        self.count: int = 0
        self.sum: float = 0.0

    def push(self, val: float):
        self.sum += val
        self.count += 1
        
    def pop(self) -> Optional[float]:
        if self.count == 0:
            return None
        r = self.sum / self.count
        self.count = 0
        self.sum = 0.0
        return r


class SystemMetricsReporter:

    def __init__(self, logger: EventLogger, report_interval: float, measure_interval: float, initial_wait: float = 3):
        assert report_interval >= 1
        assert measure_interval >= 1
        assert initial_wait >= 0
        assert report_interval >= measure_interval

        self.logger = logger
        self.report_interval = report_interval
        self.measure_interval = measure_interval
        self.initial_wait = initial_wait

        self.gputil_success = True
        self.next_measure_ts = time.perf_counter()
        self.next_report_ts = time.perf_counter()
        self.cpu_util_avg = Avarager()
        self.mem_util_avg = Avarager()
        self.gpu_util_avg = Avarager()
        self.gpu_mem_util_avg = Avarager()
        
        self.alive = True
        self.worker = Thread(target=self._worker_run, daemon=True)
        self.worker.start()


    def close(self):
        self.alive = False


    def __del__(self):
        self.close()


    def _worker_run(self):
        # wait a number of seconds before first reporting, psutil.cpu_percent() will 
        # return a meaningless value if execute immediately after its importing 
        time.sleep(self.initial_wait)

        while self.alive:
            try:
                if self.next_measure_ts <= self.next_report_ts:
                    # try measure process
                    now = time.perf_counter()
                    if now >= self.next_measure_ts:
                        self.next_measure_ts = max(now, self.next_measure_ts + self.measure_interval)
                        self._measure_process()
                else:
                    # try report process
                    now = time.perf_counter()
                    if now >= self.next_report_ts:
                        self.next_report_ts = max(now, self.next_report_ts + self.report_interval)
                        self._report_process()

                sleep_time = min(self.next_report_ts, self.next_measure_ts) - time.perf_counter()
                if sleep_time > 0:
                    time.sleep(sleep_time)

            except Exception as ex:
                # unexpected error
                self.logger.errorcf('', -1, ex, f'{EVENT_KEY_PREFIX}: unexpected error in _worker_run')
                time.sleep(3)


    def _measure_process(self):
        # cpu
        cpu_percent = psutil.cpu_percent()
        self.cpu_util_avg.push(cpu_percent)

        # memory
        memory_percent = psutil.virtual_memory().percent
        self.mem_util_avg.push(memory_percent)

        # gpu
        # Only for the first time (partial first time) unable to get gpu info, log
        # will be pushed. To avoid frequently being disturbed by cpu error/warn logs.
        gpu_ex = None
        gpu_summary = None
        try:
            gpu_summary = get_gpu_summary()
        except Exception as ex:
            gpu_ex = ex

        if gpu_ex:
            # gpu error, only push log when succeeded last time
            if self.gputil_success:
                self.gputil_success = False
                self.logger.errorcf('', -1, gpu_ex, f'{EVENT_KEY_PREFIX}: failed to get gpu info')

        elif gpu_summary is None:
            # no gpu, only push log when succeeded last time
            if self.gputil_success:
                self.gputil_success = False
                self.logger.warnf(f'{EVENT_KEY_PREFIX}: no gpu available')

        else:
            # success
            self.gputil_success = True
            self.gpu_util_avg.push(gpu_summary.utilization)
            self.gpu_mem_util_avg.push(gpu_summary.memory_utilization)


    def _report_process(self):
        # cpu
        cpu_util_val = self.cpu_util_avg.pop()
        if cpu_util_val is not None:
            self.logger.event(f'{EVENT_KEY_PREFIX}_CpuUtilization', '', cpu_util_val)
        
        # memory
        mem_util_val = self.mem_util_avg.pop()
        if mem_util_val is not None:
            self.logger.event(f'{EVENT_KEY_PREFIX}_MemoryUtilization', '', mem_util_val)
        
        # gpu
        gpu_util_val = self.gpu_util_avg.pop()
        if gpu_util_val is not None:
            self.logger.event(f'{EVENT_KEY_PREFIX}_GpuUtilization', '', gpu_util_val)
            
        # gpu mem
        gpu_mem_util_val = self.gpu_mem_util_avg.pop()
        if gpu_mem_util_val is not None:
            self.logger.event(f'{EVENT_KEY_PREFIX}_GpuMemoryUtilization', '', gpu_mem_util_val)




import sys

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from dataclasses_json import config, dataclass_json
from datetime import datetime
from enum import Enum
import threading
import traceback
from typing import Any, List, Optional


ERR_PREFIX = "CRITICAL: UNEXPECTED LOGGING ISSUE! (This logline appears only in stderr). Detail: "
LOGGING_QUEUE_CAPACITY = 10000
SINK_BATCH_MAX_SIZE = 1000
SINK_BATCH_MAX_IDLE_SECS = 5


class LogLevel(str, Enum):
    INFO = 'INFO'
    WARN = 'WARN'
    ERROR = 'ERROR'
    FATAL = 'FATAL'


class Event(ABC):
    @abstractmethod
    def friendly_string(self) -> str:
        pass
    
    @abstractmethod
    def serialize_json(self) -> str:
        pass


class TerminateEvent(Event):
    def friendly_string(self) -> str:
        return ''
    
    def serialize_json(self) -> str:
        return ''


class EventSink(ABC):
    @abstractmethod
    def sink(self, evt: Event):
        pass

    def start(self):
        pass
    
    def close(self):
        pass


class EventLogger(ABC):
    def __init__(self):
        self.mutex = threading.RLock()
        self.unstruct_sinks: List[EventSink] = []
        self.struct_sinks: List[EventSink] = []

    def start(self):
        pass

    # Structured events
    @abstractmethod
    def event(self, key: str, code: str, numeric: float, detail: str='', corr_id: str='', elem: int=-1):
        pass

    # Unstructured events
    @abstractmethod
    def log(self, level: LogLevel, msg: str, corr_id: str='', elem: int=-1):
        pass

    def infof(self, format: str, *args: Any):
        msg = format % tuple(args) if args else format
        self.log(LogLevel.INFO, msg)

    def infocf(self, corr_id: str, elem: int, format: str, *args: Any):
        msg = format % tuple(args) if args else format
        self.log(LogLevel.INFO, msg, corr_id=corr_id, elem=elem)

    def warnf(self, format: str, *args: Any):
        msg = format % tuple(args) if args else format
        self.log(LogLevel.WARN, msg)

    def warncf(self, corr_id: str, elem: int, format: str, *args: Any):
        msg = format % tuple(args) if args else format
        self.log(LogLevel.WARN, msg, corr_id=corr_id, elem=elem)

    def errorf(self, format: str, *args: Any):
        msg = format % tuple(args) if args else format
        self.log(LogLevel.ERROR, msg)

    def errorcf(self, corr_id: str, elem: int, ex: Optional[Exception], format: str, *args: Any):
        msg = format % tuple(args) if args else format
        if ex:
            msg += '\n' + get_format_ex(ex, limit=4)
        self.log(LogLevel.ERROR, msg, corr_id=corr_id, elem=elem)

    def fatalf(self, format: str, *args: Any):
        msg = format % tuple(args) if args else format
        self.log(LogLevel.FATAL, msg)

    def fatalcf(self, corr_id: str, elem: int, ex: Optional[Exception], format: str, *args: Any):
        msg = format % tuple(args) if args else format
        if ex:
            msg += '\n' + get_format_ex(ex, limit=4)
        self.log(LogLevel.FATAL, msg, corr_id=corr_id, elem=elem)

    # sink register
    def add_sink_unstructured(self, sink: EventSink):
        with self.mutex:
            self.unstruct_sinks.append(sink)        
    
    def add_sink_structured(self, sink: EventSink):
        with self.mutex:
            self.struct_sinks.append(sink)


@dataclass_json
@dataclass
class UnstructuredEvent(Event):
    level: LogLevel     = field(metadata=config(field_name='Level'))
    role: str           = field(metadata=config(field_name='Role'))
    instance: str       = field(metadata=config(field_name='Instance'))
    context: str        = field(metadata=config(field_name='Context'))
    message: str        = field(metadata=config(field_name='Message'))
    correlation_id: str = field(metadata=config(field_name='CorrelationId'))
    element: int        = field(metadata=config(field_name='Element'))
    timestamp: datetime = field(
            default_factory=datetime.utcnow,
            metadata=config(field_name='Timestamp', encoder=lambda x: datetime.isoformat(x, sep=' '))
        )
    
    def friendly_string(self) -> str:
        corr_suffix = ''
        if self.correlation_id:
            corr_suffix = "  {CorrId: %s, Elem: %d}" % (self.correlation_id, self.element)
            
        return "%s  %s(%s)  %s  %s]  %s%s\n" % (
            self.timestamp.isoformat(sep=' '), 
            self.role, self.instance,
            self.level,
            self.context, self.message,
            corr_suffix
        )
            
    def serialize_json(self) -> str:
        return self.to_json()  # type: ignore


@dataclass_json
@dataclass
class StructuredEvent(Event):
    key: str            = field(metadata=config(field_name='Key'))
    role: str           = field(metadata=config(field_name='Role'))
    instance: str       = field(metadata=config(field_name='Instance'))
    correlation_id: str = field(metadata=config(field_name='CorrelationId'))
    element: int        = field(metadata=config(field_name='Element'))
    code: str           = field(metadata=config(field_name='Code'))
    numeric: float      = field(metadata=config(field_name='Numeric'))
    detail: str         = field(metadata=config(field_name='Detail'))
    timestamp: datetime = field(
            default_factory=datetime.utcnow,
            metadata=config(field_name='Timestamp', encoder=lambda x: datetime.isoformat(x, sep=' '))
        )

    def friendly_string(self) -> str:
        corr_suffix = ''
        if self.correlation_id:
            corr_suffix = "  {CorrId: %s, Elem: %d}" % (self.correlation_id, self.element)
            
        return "%s  %s(%s)  %s:  { %s %f %s }%s\n" % (
            self.timestamp.isoformat(sep=' '), 
            self.role, self.instance,
            self.key, self.code, self.numeric, self.detail,
            corr_suffix
        )
        
    def serialize_json(self) -> str:
        return self.to_json()  # type: ignore
    
    
def get_format_ex(ex: BaseException, limit: Optional[int] = None) -> str:
    ''' limit: limit of stack frames number
    '''
    _limit = -limit if limit is not None and limit > 0 else None
    tbstr = ''.join(traceback.format_tb(ex.__traceback__, limit = _limit))
    stack = f'{type(ex).__name__}: {ex}\nTraceback\n{tbstr}'
    return stack.strip()


def _get_curr_format_ex() -> Optional[str]:
    _, ex, _ = sys.exc_info()
    if ex:
        return get_format_ex(ex)
    else:
        return None
    


import os
import socket
import threading
from typing import Optional
import uuid

from azure.identity import ManagedIdentityCredential

from .eventhub import EventHubSink
from .log import EventLogger
from .logimp import AsyncEventLogger, StdoutSink
from .reporter import SystemMetricsReporter

__all__ = [
    "infof", 
    "infocf", 
    "warnf", 
    "warncf", 
    "errorf", 
    "errorcf", 
    "fatalf",
    "fatalcf",
    "event",
    "initialize",
]

# create logger
_logger: EventLogger = AsyncEventLogger()


# logger interface
infof = _logger.infof
infocf = _logger.infocf
warnf = _logger.warnf
warncf = _logger.warncf
errorf = _logger.errorf
errorcf = _logger.errorcf
fatalf = _logger.fatalf
fatalcf = _logger.fatalcf
event = _logger.event


# logger init
_logger_init_lock = threading.Lock()
_logger_initialized: bool = False
_sys_metrics_reporter: Optional[SystemMetricsReporter] = None


def initialize(
    eh_hostname: Optional[str] = None,
    client_id: Optional[str] = None,
    eh_conn_str: Optional[str] = None,
    eh_structured: Optional[str] = None,
    eh_unstructured: Optional[str] = None,
    role: Optional[str] = None,
    instance: Optional[str] = None,
    sys_metrics_enable: bool = True,
):
        '''
        Args:
            eh_hostname: Fully Qualified Namespace aka EH Endpoint URL (*.servicebus.windows.net). Default, read `${EVENTHUB_NAMESPACE}.servicebus.windows.net`
            client_id: client_id of service principal. Default, read $UAI_CLIENT_ID
            eh_conn_str: connection string of eventhub namespace. Default, read $EVENTHUB_CONN_STRING
            eh_structured: structured eventhub name. Default, read $EVENTHUB_AUX_STRUCTURED
            eh_unstructured: unstructured eventhub name. Default, read $EVENTHUB_AUX_UNSTRUCTURED
            role: role, Default: `RemoteModel`
            instance: instance, Default: `${MODEL_NAME}|${ENDPOINT_VERSION}|{hostname}` or `${MODEL_NAME}|${ENDPOINT_VERSION}|{_probably_unique_id()}`
            sys_metrics_enable: Whether to enable auto metrics reporting periodically for system info like gpu, memory and gpu. Default: True
        
        Note: 
            1. either (eh_hostname, eh_structured or eh_unstructured) 
               or (eh_conn_str, eh_structured or eh_unstructured) is provided, 
               event hub sink will be added
            2. If client_id is not provided in any way, will get credential without specifying client_id 
               if necessary. It's helpful for testing aml endpoint without user assigned identity.
               Only need to authorize eventhub to aml endpoint.
        '''
        global _logger_initialized
        global _sys_metrics_reporter
        with _logger_init_lock:
            # skip
            if _logger_initialized:
                return
        
            # set default value
            if eh_hostname is None:
                eh_namespace = os.getenv('EVENTHUB_NAMESPACE')
                if eh_namespace:
                    eh_hostname = f'{eh_namespace}.servicebus.windows.net'
            if client_id is None:
                client_id = os.getenv('UAI_CLIENT_ID')
            if eh_conn_str is None:
                eh_conn_str = os.getenv('EVENTHUB_CONN_STRING')
            if eh_structured is None:
                eh_structured = os.getenv('EVENTHUB_AUX_STRUCTURED')
            if eh_unstructured is None:
                eh_unstructured = os.getenv('EVENTHUB_AUX_UNSTRUCTURED')
            if role is None:
                role = 'RemoteModel'
            if instance is None:
                model_name = os.getenv('MODEL_NAME', 'unknown')
                endpoint_version = os.getenv('ENDPOINT_VERSION', 'unknown')
                # Using `socket.gethostname()` to get hostname, because `os.uname` doesn't work on windows, 
                # and will be truncated in some systems. Even official doc of `os.uname` recommends to use 
                # `socket.gethostname()`. https://docs.python.org/3.5/library/os.html#os.uname
                hostname = socket.gethostname()
                if len(hostname) <= 60:
                    instance = f'{model_name}|{endpoint_version}|{hostname}'
                else:
                    instance = f'{model_name}|{endpoint_version}|{_probably_unique_id()}'

            
            # set variables
            _logger.role = role
            _logger.instance = instance
            
            # event hub sink
            if eh_conn_str:
                if eh_structured:
                    structured_sink = EventHubSink(conn_str=eh_conn_str, name=eh_structured)
                    _logger.add_sink_structured(structured_sink)
                if eh_unstructured:
                    unstructured_sink = EventHubSink(conn_str=eh_conn_str, name=eh_unstructured)
                    _logger.add_sink_unstructured(unstructured_sink)
                
            elif eh_hostname:
                if client_id:
                    credential = ManagedIdentityCredential(client_id=client_id)
                else:
                    credential = ManagedIdentityCredential()
                if eh_structured:
                    structured_sink = EventHubSink(hostname=eh_hostname, credential=credential, name=eh_structured)
                    _logger.add_sink_structured(structured_sink)
                if eh_unstructured:
                    unstructured_sink = EventHubSink(hostname=eh_hostname, credential=credential, name=eh_unstructured)
                    _logger.add_sink_unstructured(unstructured_sink)
                
            else:
                print('WARNING: logger eventhub sink is disabled')
        
            # stdout sink
            _logger.add_sink_structured(StdoutSink())
            _logger.add_sink_unstructured(StdoutSink())

            # start
            _logger.start()

            # system auto metrics
            if sys_metrics_enable:
                _sys_metrics_reporter = SystemMetricsReporter(_logger, report_interval=60, measure_interval=5)

            # success
            _logger_initialized = True
        
                
def _probably_unique_id() -> str:
	u = str(uuid.uuid4())
	return "%s-%s%s" % (u[0:5], u[5:8], u[9:11])




import queue
import time
import traceback
import multiprocessing as mp
from typing import Any, Dict, List, Optional
from dataclasses import dataclass
from threading import Thread, Condition, Event
import uuid
from flask import has_request_context, request

from pyraisdk import rlog
from .base import BaseModel
from .config import BatchingConfig


EVENT_KEY_PREFIX = 'pyraidynb'


@dataclass
class MsgSend:
    id: str
    item: Any


@dataclass
class MsgRecv:
    id: str
    res: Any
    ex: Optional[Exception]
    

class ItemFuture:
    def __init__(self, uid: str, cancellation_token: Optional[Event]=None):
        self._uid = uid
        self._condition = Condition()
        self._flag_done = False
        self._flag_cancelled = False
        self._external_ct = cancellation_token
        self._exception: Optional[Exception] = None
        self._result: Any = None
    
    def get_result(self, timeout: Optional[float]=None) -> Any:
        ''' Return the result of task future
            Args:
                timeout: the number of seconds to wait for result
                         raise PredictTimeoutError on timeout
                         Default: None, will never timeout
            Raises:
                PredictTimeoutError: If the future is not done before timeout
                PredictCancelledError: If the future is cancelled
        '''
        with self._condition:
            if self._flag_cancelled:
                raise PredictCancelledError('predict cancelled')
            if not self._flag_done:
                self._condition.wait(timeout)
                if self._flag_cancelled:
                    raise PredictCancelledError('predict cancelled')
                if not self._flag_done:
                    raise PredictTimeoutError('predict timeout')
            # from here, _flag_done == True
            if self._exception:
                raise self._exception
            return self._result

    def set_cancelled(self):
        ''' Try to mark cancelled for the task future. Won't actuallly 
            interrupt the process if the task has already started running
        '''
        with self._condition:
            if self._flag_done or self._flag_cancelled:
                return
            self._flag_cancelled = True
            self._condition.notify_all()
            
    def is_cancelled(self) -> bool:
        ''' Return if future is cancelled
        '''
        with self._condition:
            return self._flag_cancelled
        
    def is_external_ct_set(self) -> bool:
        ''' Return if external cancallation token is set
        '''
        return self._external_ct is not None and self._external_ct.is_set()

    def set_result(self, result: Any):
        ''' Set result for the future and mark it done.
            Should only be used by Promise holder
        '''
        with self._condition:
            if self._flag_done:
                return
            self._result = result
            self._flag_done = True
            self._condition.notify_all()
    
    def set_excepted(self, e: Exception):
        ''' Set exception for the future and mark it done.
            Should only be used by Promise holder
        ''' 
        with self._condition:
            if self._flag_done:
                return
            self._exception = e
            self._flag_done = True
            self._condition.notify_all()


    
@dataclass
class RequestCorrelation:
    CorrelationId: str
    Element: int


class PredictCancelledError(Exception):
    pass


class PredictTimeoutError(Exception):
    pass


class DynamicBatchModel:
    ''' Wrapped Model for dynamic batching automatically.
    
    A backend thread will serve all the callers. It merges data 
    inputs from multiple calls, and process batchly. Then return 
    corresponding results to each callers.
    
    Attributes:
        max_batch_size (int): Max size of each processing batch. Mandatory, read from environment 
                variable `PYRAISDK_MAX_BATCH_SIZE`.
        idle_batch_size (int): If there's no more data in queue, a new batch will be launched when
                size reaches this value. Mandatory, read from environment variable `PYRAISDK_IDLE_BATCH_SIZE`.
        max_batch_interval(float): Max interval in seconds to wait for items. When waiting time 
                exceeds, will launch a batch immediately. Mandatory, read from environment variable 
                `PYRAISDK_MAX_BATCH_INTERVAL`.

    '''
    
    def __init__(self, model: BaseModel):
        '''
        Args: 
            model: Subclass of BaseMode, to provide business logic

        '''
        # read batching config
        batching_cfg: BatchingConfig = BatchingConfig.populate()
        self.max_batch_size: int = batching_cfg.pyraisdk_max_batch_size
        self.idle_batch_size: int = batching_cfg.pyraisdk_idle_batch_size
        self.max_batch_interval: float = batching_cfg.pyraisdk_max_batch_interval

        assert self.max_batch_size > 0, f'Invalid PYRAISDK_MAX_BATCH_SIZE value {self.max_batch_size}'
        assert self.idle_batch_size > 0, f'Invalid PYRAISDK_IDLE_BATCH_SIZE value {self.idle_batch_size}'
        assert self.max_batch_interval > 0.0, f'Invalid MAX_BATCH_INTERVAL value {self.max_batch_interval}'
        assert self.idle_batch_size <= self.max_batch_size, f'Require PYRAISDK_IDLE_BATCH_SIZE ({self.idle_batch_size}) <= PYRAISDK_MAX_BATCH_SIZE ({self.max_batch_size})'

        self.model = model
        self.futuredict: Dict[str, ItemFuture] = {}
        self.alive = mp.Event()
        self.q_send: mp.Queue[MsgSend] = mp.Queue()
        self.q_recv: mp.Queue[MsgRecv] = mp.Queue()
        
        self.worker = mp.Process(target=self._worker_run, daemon=True)
        self.worker.start()
        self.reciever = Thread(target=self._reciever_run, daemon=True)
        self.reciever.start()
    
    def predict(
        self,
        items: List[Any],
        timeout: Optional[float] = 1.0,
        raise_timeout: bool = True,
        cancellation_token: Optional[Event] = None,
        req_corr: Optional[RequestCorrelation] = None,
    ) -> List[Any]:
        ''' Predict with dynamic batching
        
        Args:
            items: model input list
            timeout: A floating point number specifying a timeout for the operation in seconds, or None 
                    indicating to wait without timeout
            raise_timeout: When timeout, True: raise PredictTimeoutError; False: set to None for 
                    no result items and return.
            cancellation_token: A event for cancellation. If the event is set, will try to cancel and skip
                    the calculation in downstream batch process. Default: None.
            req_corr: Request correlation struct composed of request's Correlation ID and Element. Default 
                    value composed from request headers "Correlation-Id" and "Element"
            
        Returns:
            predict output list
        
        '''
        if req_corr is None:
            req_corr = get_request_correlation()

        ts_start = time.perf_counter()
        try:
            # get futures & put messsage into queue
            futurelist: List[ItemFuture] = []
            for item in items:
                uid = str(uuid.uuid4())
                future = ItemFuture(uid, cancellation_token)
                self.futuredict[uid] = future
                msg = MsgSend(uid, item)
                self.q_send.put_nowait(msg)
                futurelist.append(future)
            
            # get results
            rs = []
            start_time = time.perf_counter()
            for future in futurelist:
                if timeout is None:
                    time_left = None
                else:
                    time_left = timeout - (time.perf_counter() - start_time)
                try:
                    result = future.get_result(timeout=time_left)
                except PredictTimeoutError:
                    future.set_cancelled()
                    try:
                        del self.futuredict[future._uid]
                    except KeyError:
                        pass
                    if raise_timeout:
                        raise
                    else:
                        result = None
                rs.append(result)
                
            # duration
            rlog.event(f'{EVENT_KEY_PREFIX}_PredictDuration', 'success', time.perf_counter() - ts_start, corr_id=req_corr.CorrelationId, elem=req_corr.Element)
            return rs
        
        except Exception as ex:
            rlog.event(f'{EVENT_KEY_PREFIX}_PredictDuration', 'error', time.perf_counter() - ts_start, corr_id=req_corr.CorrelationId, elem=req_corr.Element)
            rlog.errorcf(req_corr.CorrelationId, req_corr.Element, ex, f'{EVENT_KEY_PREFIX}: error in predict')
            raise


    def predict_one(
        self,
        item: Any,
        timeout: Optional[float] = 1.0,
        raise_timeout: bool = True,
        cancellation_token: Optional[Event] = None,
        req_corr: Optional[RequestCorrelation] = None,
    ) -> Any:
        ''' Predict for a single item, with dynamic batching
        
        Args:
            item: model input, a single item
            timeout: A floating point number specifying a timeout for the operation in seconds, or None 
                    indicating to wait without timeout
            raise_timeout: When timeout, True: raise PredictTimeoutError; False: set to None for 
                    no result items and return.
            cancellation_token: A event for cancellation. If the event is set, will try to cancel and skip
                    the calculation in downstream batch process. Default: None.
            req_corr: Request correlation struct composed of request's Correlation ID and Element. Default 
                    value composed from request headers "Correlation-Id" and "Element"
            
        Returns:
            single predict output
        
        '''
        rs = self.predict(
            [item], 
            timeout=timeout, 
            raise_timeout=raise_timeout, 
            cancellation_token=cancellation_token,
            req_corr=req_corr,
        )
        return rs[0]

    
    def close(self):
        self.alive.set()
        
        
    def __del__(self):
        self.close()
        
        
    def _worker_run(self):
        ''' If _worker_inner failed somehow, sleep some time and retry
        '''
        self.model.load()
        while not self.alive.is_set():
            try:
                self._worker_inner()
            except Exception:
                # unexpected error
                traceback.print_exc()
                time.sleep(0.1)
                
                
    def _worker_inner(self):
        ''' condition to trigger a batch run:
            1. [batch size] >= self.max_batch_size
            2. queue is empty && [batch size] >= self.idle_batch_size
            3. [now] - [last run time] >= self.max_batch_interval
        '''
        batch: List[MsgSend] = []          # cached batch items
        last_run_time = time.perf_counter()    # last batch run time
        
        while not self.alive.is_set():
            # get from q without block & check empty
            try:
                msg = self.q_send.get_nowait()
                batch.append(msg)
            except queue.Empty:
                msg = None

            # here msg is None means q is empty, idle
            # idle && size >= idle_batch_size, trigger batch run
            if msg is None and len(batch) >= self.idle_batch_size:
                self._worker_run_batch(batch)
                batch = []
                last_run_time = time.perf_counter()
                continue

            # q is empty and still have time left, get from q with timeout
            if msg is None:
                timeout = self.max_batch_interval - (time.perf_counter() - last_run_time)
                if timeout > 0:
                    try:
                        msg = self.q_send.get(block=True, timeout=timeout)
                        # handle cancellation
                        batch.append(msg)
                    except queue.Empty:
                        msg = None

            # here msg is None means timeout
            # timeout or batch reached max size, trigger batch run
            if msg is None or len(batch) >= self.max_batch_size:
                if batch:
                    self._worker_run_batch(batch)
                    batch = []
                last_run_time = time.perf_counter()
            
            
    def _worker_run_batch(self, batch: List[MsgSend]):
        # preprocess & predict
        batch_track_id = str(uuid.uuid4())
        try:
            input = [msg.item for msg in batch]
            rlog.event(f'{EVENT_KEY_PREFIX}_BatchSize', '', len(batch), corr_id=batch_track_id)
                
            ts_start = time.perf_counter()
            try:
                output_preprocess = self.model.preprocess(input)
            except:
                rlog.event(f'{EVENT_KEY_PREFIX}_BatchPreprocessDuration', 'error', time.perf_counter() - ts_start, corr_id=batch_track_id)
                raise
            else:
                rlog.event(f'{EVENT_KEY_PREFIX}_BatchPreprocessDuration', 'success', time.perf_counter() - ts_start, corr_id=batch_track_id)

            try:
                output_predict = self.model.predict(output_preprocess)
            except:
                rlog.event(f'{EVENT_KEY_PREFIX}_BatchDuration', 'error', time.perf_counter() - ts_start, corr_id=batch_track_id)
                raise
            else:
                rlog.event(f'{EVENT_KEY_PREFIX}_BatchDuration', 'success', time.perf_counter() - ts_start, corr_id=batch_track_id)
            
            if len(output_predict) != len(batch):
                raise Exception(f'model output size is {len(output_predict)} while input size is {len(batch)}')

        except Exception as e:
            for msg in batch:
                mrecv = MsgRecv(msg.id, None, e)
                self.q_recv.put_nowait(mrecv)
                
            rlog.errorcf(batch_track_id, -1, e, f'{EVENT_KEY_PREFIX}: error running a batch')
            # items avg latency
            ts_end = time.perf_counter()
            #latency_avg = sum(ts_end - msg.create_ts for msg in batch) / len(batch)
            #rlog.event(f'{EVENT_KEY_PREFIX}_BatchItemsAvgLatency', 'error', latency_avg, corr_id=batch_track_id)
            return
        
        # set result for each
        for msg, result in zip(batch, output_predict):
            mrecv = MsgRecv(msg.id, result, None)
            self.q_recv.put_nowait(mrecv)            

        # items avg latency
        ts_end = time.perf_counter()
        #latency_avg = sum(ts_end - msg.create_ts for msg in batch) / len(batch)
        #rlog.event(f'{EVENT_KEY_PREFIX}_BatchItemsAvgLatency', 'success', latency_avg, corr_id=batch_track_id)

    def _reciever_run(self):
        while not self.alive.is_set():
            try:
                self._reciever_run_inner()
            except Exception:
                # unexpected error
                traceback.print_exc()
                time.sleep(1)
    
    def _reciever_run_inner(self):
        while True:
            msg: MsgRecv = self.q_recv.get()
            future = self.futuredict.get(msg.id)
            if future is None:
                continue
            if msg.ex:
                future.set_excepted(msg.ex)
            else:
                future.set_result(msg.res)


def get_request_correlation() -> RequestCorrelation:
    ''' get Correlation ID and Element from flask http headers
    '''
    if has_request_context():
        return RequestCorrelation(
            request.headers.get('Correlation-Id', ''),
            int(request.headers.get('Element', '-1')),
        )
    else:
        return RequestCorrelation('', -1)



from abc import ABC, abstractmethod
from typing import Any, List

class BaseModel(ABC):
    '''Base class for models to inherit from
    '''
    
    def load(self):
        pass

        
    @abstractmethod
    def predict(self, items: List[Any]) -> List[Any]:
        '''Model predict function
        
        Abstract method which need to be implemented in subclass. It receives
        a bunch of input items, and output same number of predict results.
        
        Args:
            items: Input data items
            
        Returns:
            Predict results for each input data item
        '''
        pass
        
        
    def preprocess(self, items: List[Any]) -> List[Any]:
        '''Preprocessing steps to convert raw data into the form required by your model
        
        Put your precessing logic into this function so that metrics like counting
        and duration will be handled correctly. For example, perform any JSON transformations
        you need here

        Default implementation is enough if you just want to put every thing in the 'predict' function.
        
        Args:
            items: Input data items
            
        Returns:
            Preprocessing results for each input data item
        '''
        return items
    



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyraisdk', 'pyraisdk.dynbatch', 'pyraisdk.rlog']

package_data = \
{'': ['*']}

install_requires = \
['azure-eventhub>=5.10.1,<6.0.0',
 'azure-identity>=1.7.0,<2.0.0',
 'dataclasses-json>=0.5.7,<0.6.0',
 'flask',
 'psutil>=5.6.3,<6.0.0']

entry_points = \
{'console_scripts': ['publish_if_not_exists = '
                     'scripts.publish_if_not_exists:main']}

setup_kwargs = {
    'name': 'pyraisdk-test',
    'version': '0.2.5',
    'description': '',
    'long_description': '# pyraisdk\n\n## Overview\n\nAML models are meant to be deployed to GPU instances to provide inference service. If the code that operates the model uses the GPU for inferencing in each request separately,\nthe overall performance of the model will be quite inefficent. This SDK has APIs that can allocate batches of inference requests to run on the GPU in a separate thread, thereby considerably\nimproving the usage efficiency of GPU and making the model more performant. \n\nThe SDK also collects telemetry data for each so the performance of the model can be evaluated and tracked and provides logging primitives that can be used\nto produce additional troubleshooting information.\n\n\n## Using pyraisdk\n\n### Integrating with Model\nFollow the steps below to integrate pyraisdk.\n\n1. In environment configuration, list latest [pyraisdk version](https://github.com/microsoft/rai-orchestrator/blob/master/pyraisdk/pyproject.toml#L3) as a dependency.\n2. Import resources from pyraisdk\n  ```\n  from pyraisdk import rlog\n  from pyraisdk.dynbatch import BaseModel, DynamicBatchModel\n  ```\n3. Implement `BaseModel` class, which includes methods `preprocess()` and `predict()`, to define how data must be preprocessed and how inferences are made.\n4. At process start (e.g. in `init()` function if deploying with scoring script), initialize logging, instantiate object of class that implements `BaseModel`, and define global variable for batched model as shown below. \n  ```\n  rlog.initialize()\n  \n  global batched_model\n  malware_model = MalwareModel()\n  batched_model = DynamicBatchModel(malware_model)\n  ```\n5. In request handler (e.g. in `run()` function if deploying with scoring script), pass list of requests to the batched model\'s `predict()` method and return results. Under-the-hood, pyraisdk creates batches and uses your model\'s `preprocess()` and `predict()` methods to generate results.\n  ```\n  return batched_model.predict(json.loads(request_data)[\'data\'])\n  ```\n6. Optionally log additional structured and unstructured events using `EventLogger` [methods](https://github.com/microsoft/rai-orchestrator/blob/master/pyraisdk/pyraisdk/rlog/log.py). It is not necessary to log latency information, as pyraisdk does so automatically.\n\n### Configuring CD Pipeline\nIn addition to following the CD pipeline\'s [developer guidelines](https://github.com/microsoft/rai-orchestrator/blob/master/cicd/cd/regional/raicd/README.md#prerequisites-to-adding-or-modifying-endpoint-configurations) to onboard your model to RAI\'s deployment infrastructure, you must configure `BatchingConfig` for each deployment target to which your model will be deployed. To do so, \n1. Open the appropriate configuration file under `deployment-target-configs/../*.json`\n2. Configure the model\'s `Version`, `InstanceType` / SKU, and `BatchingConfig` as shown below\n  ```json\n  {\n\t  "Name": "MalwareNeural",\n\t  "Version": 19,\n      "InstanceType": "Standard_NC6s_v3",\n      "BatchingConfig": {\n          "MaxBatchSize": 12,\n          "IdleBatchSize": 5,\n          "MaxBatchInterval": 0.002\n      }\n  }\n  ```\nWhen the CD pipeline deploys the model to an AML online deployment, it exports environment variables for each `BatchingConfig` field, which pyraisdk utilizes at process start to configure batching parameters. The CD pipeline also exports environment variables to enable pyraisdk to log to RAI\'s Azure Data Explorer clusters.\n\n### Test or Deploy without RAI CD Pipeline\n\nTest or Deploy independetly outside of RAI CD Pipeline, it\'s required to set environment variables for `BatchingConfig` manually before object of `DynamicBatchModel` is created. Refer to [Batching Parameter](#batching-parameter-attention). \n\nThe recommended way to set these variables is calling `os.environ.setdefault` in `init()`. Like below:\n\n```python\ndef init():\n    os.environ.setdefault(\'PYRAISDK_MAX_BATCH_SIZE\', \'12\')\n    os.environ.setdefault(\'PYRAISDK_IDLE_BATCH_SIZE\', \'5\')\n    os.environ.setdefault(\'PYRAISDK_MAX_BATCH_INTERVAL\', \'0.002\')\n    ...\n    batch_model = DynamicBatchModel(malware_model)\n```\n\nThis should work well independent of RAI CD Pipeline. And you don\'t need to specifically remove these 3 lines when switching to deploy in CD Pipeline. \n\nBut please note that this kind of code should be **Avoided**: `os.environ[\'PYRAISDK_MAX_BATCH_SIZE\'] = \'12\'`. It will override the configuration from CD pipeline, which is not desirable in most cases.\n\nTo enable log publishing (to eventhub), there are another several environment variables need to be set, refer to following logging part. It\'s optional.\n\n\n## Dynamic Batching Support\n\nThere are APIs you **must** implement in your model to support batching of inference requests for best model performance. Those APIs allow the SDK\nto distribute load efficiently to the GPU instances. The APIs are:\n\n* ```preprocess``` Modifies the input to the model, if necessary. For example, if your model needs the input in a special JSON format instead of as a list\nof strings, you can do that modification in the *preprocess* method.\n* ```predict``` Executes the model inference for a list of input strings\n\n### Batching Parameter (**Attention**)\n\nBatching parameters are mandatory and should be ready before DynamicBatchModel is created. They are set through environment variables:\n- **PYRAISDK_MAX_BATCH_SIZE** (int): Max size of each processing batch.\n- **PYRAISDK_IDLE_BATCH_SIZE** (int): If there\'s no more data in queue, a new batch will be launched when size reaches this value.\n- **PYRAISDK_MAX_BATCH_INTERVAL** (float): Max interval in seconds to wait for items. When waiting time exceeds, will launch a batch immediately.\n\n\n### Usage Examples\n\nBuild `YourModel` class inherited from `pyraisdk.dynbatch.BaseModel`.\n\n```python\nfrom typing import List\nfrom pyraisdk.dynbatch import BaseModel\n\nclass YourModel(BaseModel):\n    def predict(self, items: List[str]) -> List[int]:\n        rs = []\n        for item in items:\n            rs.append(len(item))\n        return rs\n            \n    def preprocess(self, items: List[str]) -> List[str]:\n        rs = []\n        for item in items:\n            rs.append(f\'[{item}]\')\n        return rs\n```\n\nInitialize a `pyraisdk.dynbatch.DynamicBatchModel` with `YourModel` instance, and call `predict / predict_one` for inferencing.\n\n```python\nfrom pyraisdk.dynbatch import DynamicBatchModel\n\n# prepare model\nsimple_model = YourModel()\nbatch_model = DynamicBatchModel(simple_model)\n\n# predict\nitems = [\'abc\', \'123456\', \'xyzcccffaffaaa\']\npredictions = batch_model.predict(items)\nassert predictions == [5, 8, 16]\n\n# predict_one\nitem = \'abc\'\nprediction = batch_model.predict_one(item)\nassert prediction == 5\n```\n\nConcurrent requests to `predict / predict_one`, in different threads.\n\n```python\nfrom threading import Thread\nfrom pyraisdk.dynbatch import DynamicBatchModel\n\n# prepare model\nsimple_model = YourModel()\nbatch_model = DynamicBatchModel(simple_model)\n\n# thread run function\ndef run(name, num):\n    for step in range(num):\n        item = f\'{name}-{step}\'\n        prediction = batch_model.predict_one(item)\n        assert prediction == len(item) + 2\n\n# start concurrent inference\nthreads = [Thread(target=run, args=(f\'{tid}\', 100)) for tid in range(20)]\nfor t in threads:\n    t.start()\nfor t in threads:\n    t.join()\n```\n\n## Loging & Events\n\n### Description\nThis module is for logging and event tracing.\n\n### interface\n\n```python\ndef initialize(\n    eh_hostname: Optional[str] = None,\n    client_id: Optional[str] = None,\n    eh_conn_str: Optional[str] = None,\n    eh_structured: Optional[str] = None,\n    eh_unstructured: Optional[str] = None,\n    role: Optional[str] = None,\n    instance: Optional[str] = None,\n    sys_metrics_enable: bool = True,\n)\n```\n\nParameter description for `initialize`:\n- **eh_hostname**: Fully Qualified Namespace aka EH Endpoint URL (*.servicebus.windows.net). Default, read `${EVENTHUB_NAMESPACE}.servicebus.windows.net`\n- **client_id**: client_id of service principal. Default, read $UAI_CLIENT_ID\n- **eh_conn_str**: connection string of eventhub namespace. Default, read $EVENTHUB_CONN_STRING\n- **eh_structured**: structured eventhub name. Default, read $EVENTHUB_AUX_STRUCTURED\n- **eh_unstructured**: unstructured eventhub name. Default, read $EVENTHUB_AUX_UNSTRUCTURED\n- **role**: role, Default: `RemoteModel`\n- **instance**: instance, Default: `${MODEL_NAME}|${ENDPOINT_VERSION}|{hostname}` or `${MODEL_NAME}|${ENDPOINT_VERSION}|{_probably_unique_id()}`\n- **sys_metrics_enable**: Whether to enable auto metrics reporting periodically for system info like gpu, memory and gpu. Default: True\n\n```python\ndef event(self, key: str, code: str, numeric: float, detail: str=\'\', corr_id: str=\'\', elem: int=-1)\ndef infof(self, format: str, *args: Any)\ndef infocf(self, corr_id: str, elem: int, format: str, *args: Any)\ndef warnf(self, format: str, *args: Any)\ndef warncf(self, corr_id: str, elem: int, format: str, *args: Any)\ndef errorf(self, format: str, *args: Any)\ndef errorcf(self, corr_id: str, elem: int, ex: Optional[Exception], format: str, *args: Any)\ndef fatalf(self, format: str, *args: Any)\ndef fatalcf(self, corr_id: str, elem: int, ex: Optional[Exception], format: str, *args: Any)\n```\n\n### examples\n\n```python\n# export EVENTHUB_AUX_UNSTRUCTURED=\'ehunstruct\'\n# export EVENTHUB_AUX_STRUCTURED=\'ehstruct\'\n# export UAI_CLIENT_ID=\'xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx\'\n# export EVENTHUB_NAMESPACE=\'raieusdev-eh-namespace\'\n\nfrom pyraisdk import rlog\nrlog.initialize()\n\nrlog.infof(\'this is a info message %s\', 123)\nrlog.event(\'LifetimeEvent\', \'STOP_GRACEFUL_SIGNAL\', 0, \'detail info\')\n```\n\n```python\n# export EVENTHUB_AUX_UNSTRUCTURED=\'ehunstruct\'\n# export EVENTHUB_AUX_STRUCTURED=\'ehstruct\'\n# export EVENTHUB_CONN_STRING=\'<connection string>\'\n\nfrom pyraisdk import rlog\nrlog.initialize()\n\nrlog.infocf(\'corrid\', -1, \'this is a info message: %s\', 123)\nrlog.event(\'RequestDuration\', \'200\', 0.01, \'this is duration in seconds\')\n```\n\n```python\nfrom pyraisdk import rlog\nrlog.initialize(eh_structured=\'ehstruct\', eh_unstructured=\'ehunstruct\', eh_conn_str=\'<eventhub-conn-str>\')\n\nrlog.errorcf(\'corrid\', -1, Exception(\'error msg\'), \'error message: %s %s\', 1,2)\nrlog.event(\'CpuUsage\', \'\', 0.314, detail=\'cpu usage\', corr_id=\'corrid\', elem=-1)\n```\n',
    'author': 'Xiaodong Yang',
    'author_email': 'xiaoyan@microsoft.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# refer to GPUtil
# https://github.com/anderskm/gputil/blob/master/GPUtil/GPUtil.py


from dataclasses import dataclass
import math
import os
import platform
import subprocess
from distutils import spawn
from typing import Dict, List, Optional


@dataclass
class GPUInfo:
    index: int
    uuid: str
    utilization_gpu: float
    memory_total: float
    memory_used: float
    memory_free: float
    driver_version: str
    name: str
    gpu_serial: str
    display_active: str
    display_mode: str
    temperature_gpu: float
    

@dataclass
class GPUSummary:
    count: int
    utilization: float
    memory_utilization: float


def get_gpus() -> List[GPUInfo]:
    nvidia_smi = _get_nvidia_smi()
    fields = [
        'index',
        'uuid',
        'utilization.gpu',
        'memory.total',
        'memory.used',
        'memory.free',
        'driver_version',
        'name',
        'gpu_serial',
        'display_active',
        'display_mode',
        'temperature.gpu',
    ]
    proc = subprocess.Popen([nvidia_smi, f"--query-gpu={','.join(fields)}", "--format=csv,noheader,nounits"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    try:
        outs, errs = proc.communicate(timeout=10)
    finally:
        if proc.poll() is None:
            proc.kill()

    if proc.poll() != 0:
        errstr = errs.decode('UTF-8')
        raise RuntimeError(f'nvidia-smi exit with code {proc.returncode}, error: {errstr}')
        
    outstr = outs.decode('UTF-8')
    lines = [line for line in outstr.split(os.linesep) if line.strip()]
    gpus = []
    for line in lines:
        values = [v.strip() for v in line.split(',')]
        if len(fields) != len(values):
            raise RuntimeError(f'nvidia-smi invalid output line: {line}')
        d = dict(zip(fields, values))
        g = GPUInfo(
            index = int(d['index']),
            uuid = d['uuid'],
            utilization_gpu = _get_float(d, 'utilization.gpu') / 100,
            memory_total = _get_float(d, 'memory.total'),
            memory_used = _get_float(d, 'memory.used'),
            memory_free = _get_float(d, 'memory.free'),
            driver_version = d['driver_version'],
            name = d['name'],
            gpu_serial = d['gpu_serial'],
            display_active = d['display_active'],
            display_mode = d['display_mode'],
            temperature_gpu = _get_float(d, 'temperature.gpu'),
        )
        gpus.append(g)
    return gpus


def get_gpu_summary() -> Optional[GPUSummary]:
    gpus = get_gpus()
    if not gpus:
        return None
    summary = GPUSummary(
        count = len(gpus),
        utilization = sum(g.utilization_gpu for g in gpus) / len(gpus),
        memory_utilization = sum(g.memory_used for g in gpus) / sum(g.memory_total for g in gpus),
    )
    return summary
    

def _get_float(d: Dict[str, str], name: str) -> float:
    s = d[name]
    try:
        v = float(s)
    except ValueError:
        raise RuntimeError(f'could not convert "{s}" (field {name}) to float')
    
    if not math.isfinite(v):
        raise RuntimeError(f'not finite float "{s}" (field {name})')
    return v


def _get_nvidia_smi() -> str:
    if platform.system() == "Windows":
        # If the platform is Windows and nvidia-smi 
        # could not be found from the environment path, 
        # try to find it from system drive with default installation path
        nvidia_smi = spawn.find_executable('nvidia-smi')
        if nvidia_smi is None:
            nvidia_smi = "%s\\Program Files\\NVIDIA Corporation\\NVSMI\\nvidia-smi.exe" % os.environ['systemdrive']
    else:
        nvidia_smi = "nvidia-smi"
    return nvidia_smi
